@@ -1,6 +1,6 @@
 __title__ = "pathy"
-__version__ = "0.2.0"
-__summary__ = "pathlib.Path subclasses for Google Cloud Storage"
+__version__ = "0.3.0"
+__summary__ = "pathlib.Path subclasses for local and cloud bucket storage"
 __uri__ = "https://github.com/justindujardin/pathy"
 __author__ = "Justin DuJardin"
 __email__ = "justin@explosion.ai"
