@@ -1,24 +1,34 @@
-from dataclasses import dataclass, field
-from typing import Optional, List, Generator
-from .client import BucketClient, ClientBucket, ClientBlob, ClientError, BucketEntry
-from .base import PurePathy
+from dataclasses import dataclass
+from typing import Any, Dict, Generator, List, Optional
+
+from .base import Blob, Bucket, BucketClient, BucketEntry, ClientError, PurePathy
 
 try:
-    from google.cloud import storage
     from google.api_core import exceptions as gcs_errors
+    from google.auth.exceptions import DefaultCredentialsError
+    from google.cloud import storage
 
     has_gcs = True
 except ImportError:
     storage = None
     has_gcs = False
 
+_MISSING_DEPS = """You are using the GCS functionality of Pathy without
+having the required dependencies installed.
 
-class BucketEntryGCS(BucketEntry["ClientBucketGCS", storage.Blob]):
+Please try installing them:
+
+    pip install pathy[gcs]
+
+"""
+
+
+class BucketEntryGCS(BucketEntry["BucketGCS", "storage.Blob"]):
     ...
 
 
 @dataclass
-class ClientBlobGCS(ClientBlob["ClientBucketGCS", storage.Blob]):
+class BlobGCS(Blob["storage.Bucket", "storage.Blob"]):
     def delete(self) -> None:
         self.raw.delete()
 
@@ -27,92 +37,134 @@ class ClientBlobGCS(ClientBlob["ClientBucketGCS", storage.Blob]):
 
 
 @dataclass
-class ClientBucketGCS(ClientBucket):
+class BucketGCS(Bucket):
     name: str
-    bucket: storage.Bucket
+    bucket: "storage.Bucket"
 
-    def get_blob(self, blob_name: str) -> Optional[ClientBlobGCS]:
-        native_blob = self.bucket.get_blob(blob_name)
+    def get_blob(self, blob_name: str) -> Optional[BlobGCS]:
+        assert isinstance(
+            blob_name, str
+        ), f"expected str blob name, but found: {type(blob_name)}"
+        native_blob = None
+        try:
+            native_blob = self.bucket.get_blob(blob_name)
+        except gcs_errors.ClientError:
+            pass
         if native_blob is None:
             return None
-        return ClientBlobGCS(
+        return BlobGCS(
             bucket=self.bucket,
             owner=native_blob.owner,
             name=native_blob.name,
             raw=native_blob,
             size=native_blob.size,
-            updated=native_blob.updated.timestamp(),
+            updated=int(native_blob.updated.timestamp()),
         )
 
-    def copy_blob(
-        self, blob: ClientBlobGCS, target: "ClientBucketGCS", name: str
-    ) -> Optional[ClientBlobGCS]:
+    def copy_blob(  # type:ignore[override]
+        self, blob: BlobGCS, target: "BucketGCS", name: str
+    ) -> Optional[BlobGCS]:
         assert blob.raw is not None, "raw storage.Blob instance required"
         native_blob = self.bucket.copy_blob(blob.raw, target.bucket, name)
         if native_blob is None:
             return None
-        return ClientBlobGCS(
+        return BlobGCS(
             bucket=self.bucket,
             owner=native_blob.owner,
             name=native_blob.name,
             raw=native_blob,
             size=native_blob.size,
-            updated=native_blob.updated.timestamp(),
+            updated=int(native_blob.updated.timestamp()),
         )
 
-    def delete_blob(self, blob: ClientBlobGCS) -> None:
+    def delete_blob(self, blob: BlobGCS) -> None:  # type:ignore[override]
         return self.bucket.delete_blob(blob.name)
 
-    def delete_blobs(self, blobs: List[ClientBlobGCS]) -> None:
+    def delete_blobs(self, blobs: List[BlobGCS]) -> None:  # type:ignore[override]
         return self.bucket.delete_blobs(blobs)
 
+    def exists(self) -> bool:
+        try:
+            return self.bucket.exists()
+        except gcs_errors.ClientError:
+            return False
+
 
-@dataclass
 class BucketClientGCS(BucketClient):
-    client: storage.Client = field(default_factory=lambda: storage.Client())
+    client: Optional["storage.Client"]
 
-    def make_uri(self, path: PurePathy):
+    def __init__(self, client: Optional["storage.Client"] = None):
+        try:
+            self.client = storage.Client() if storage else None
+        except (BaseException, DefaultCredentialsError):
+            self.client = None
+
+    def make_uri(self, path: PurePathy) -> str:
         return str(path)
 
-    def create_bucket(self, path: PurePathy) -> ClientBucket:
+    def create_bucket(self, path: PurePathy) -> Bucket:
+        assert self.client is not None, _MISSING_DEPS
         return self.client.create_bucket(path.root)
 
     def delete_bucket(self, path: PurePathy) -> None:
+        assert self.client is not None, _MISSING_DEPS
         bucket = self.client.get_bucket(path.root)
         bucket.delete()
 
-    def lookup_bucket(self, path: PurePathy) -> Optional[ClientBucketGCS]:
+    def exists(self, path: PurePathy) -> bool:
+        # Because we want all the parents of a valid blob (e.g. "directory" in
+        # "directory/foo.file") to return True, we enumerate the blobs with a prefix
+        # and compare the object names to see if they match a substring of the path
+        key_name = str(path.key)
         try:
-            native_bucket = self.client.lookup_bucket(path.root)
-            if native_bucket is not None:
-                return ClientBucketGCS(str(path.root), bucket=native_bucket)
+            for obj in self.list_blobs(path):
+                if obj.name == key_name:
+                    return True
+                if obj.name.startswith(key_name + path._flavour.sep):
+                    return True
         except gcs_errors.ClientError:
-            pass
+            return False
+        return False
+
+    def lookup_bucket(self, path: PurePathy) -> Optional[BucketGCS]:
+        assert self.client is not None, _MISSING_DEPS
+        try:
+            native_bucket = self.client.bucket(path.root)
+            if native_bucket is not None:
+                return BucketGCS(str(path.root), bucket=native_bucket)
+        except gcs_errors.ClientError as err:
+            print(err)
+
         return None
 
-    def get_bucket(self, path: PurePathy) -> ClientBucketGCS:
+    def get_bucket(self, path: PurePathy) -> BucketGCS:
+        assert self.client is not None, _MISSING_DEPS
         try:
-            native_bucket = self.client.lookup_bucket(path.root)
+            native_bucket = self.client.bucket(path.root)
             if native_bucket is not None:
-                return ClientBucketGCS(str(path.root), bucket=native_bucket)
+                return BucketGCS(str(path.root), bucket=native_bucket)
             raise FileNotFoundError(f"Bucket {path.root} does not exist!")
         except gcs_errors.ClientError as e:
             raise ClientError(message=e.message, code=e.code)
 
-    def list_buckets(self, **kwargs) -> Generator[ClientBucket, None, None]:
-        return self.client.list_buckets(**kwargs)
+    def list_buckets(
+        self, **kwargs: Dict[str, Any]
+    ) -> Generator["storage.Bucket", None, None]:
+        assert self.client is not None, _MISSING_DEPS
+        return self.client.list_buckets(**kwargs)  # type:ignore
 
-    def scandir(
+    def scandir(  # type:ignore[override]
         self,
         path: Optional[PurePathy] = None,
         prefix: Optional[str] = None,
         delimiter: Optional[str] = None,
-        include_raw: bool = False,
-    ) -> Generator[BucketEntryGCS, None, None]:
+    ) -> Generator[BucketEntryGCS, None, None]:  # type:ignore[override]
+        assert self.client is not None, _MISSING_DEPS
         continuation_token = None
         if path is None or not path.root:
-            for bucket in self.list_buckets():
-                yield BucketEntryGCS(bucket.name, is_dir=True, raw=None)
+            gcs_bucket: "storage.Bucket"
+            for gcs_bucket in self.list_buckets():
+                yield BucketEntryGCS(gcs_bucket.name, is_dir=True, raw=None)
             return
         sep = path._flavour.sep
         bucket = self.lookup_bucket(path)
@@ -156,7 +208,8 @@ class BucketClientGCS(BucketClient):
         prefix: Optional[str] = None,
         delimiter: Optional[str] = None,
         include_dirs: bool = False,
-    ) -> Generator[ClientBlobGCS, None, None]:
+    ) -> Generator[BlobGCS, None, None]:
+        assert self.client is not None, _MISSING_DEPS
         continuation_token = None
         bucket = self.lookup_bucket(path)
         if bucket is None:
@@ -175,7 +228,7 @@ class BucketClientGCS(BucketClient):
                 )
             for page in response.pages:
                 for item in page:
-                    yield ClientBlobGCS(
+                    yield BlobGCS(
                         bucket=bucket,
                         owner=item.owner,
                         name=item.name,
